﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.MetadirectoryServices;
using SoftwareIDM.Uplift;

namespace SoftwareIDM.ExampleExtension
{
    static class ExampleFunctions
    {
        // IIF(IsNullOrEmpty(mv.sAMAccountName), GetSAMAccountName(givenName, sn), Skip())->sAMAccountName
        public static AttributeValue GetSAMAccountName(AttributeValue givenName, AttributeValue sn)
        {
            return new AttributeValue(givenName.StringValue.SafeSubstring(0, 1) + sn.StringValue);
        }
    }
    /// <summary>
    /// This MAExtension object can be used directly for rules extensions that don't require
    /// further custom code, but it is primarily intended as a prototypical example illustrating
    /// how to use the Parser in a Rules Extension.
    /// </summary>
    public class ExampleExtension : IMASynchronization
    {
        /// <summary>
        /// The RuleEngine is the core class for Rule based attribute flow.
        /// </summary>
        RuleEngine rules;

        // Set the location of the app.config file for configuration data
        public ExampleExtension() {
            Config.ExeConfigPath = @"C:\Program Files\Microsoft Forefront Identity Manager\2010\Synchronization Service\Extensions\App.config";

            /* 
              Example of a mapping XML file
              Config.Map.Get("Locations", "60606", "ou");
            */
            Config.Map.Add("Locations", @"C:\Program Files\Microsoft Forefront ...\Extensions\locations.xml",
                "ou", "postalCode", true);

            rules = new RuleEngine(typeof(CommonFunctions), typeof(ExampleFunctions));
            rules.ChangeHoloFields = new Dictionary<string, string[]>() { 
                { "user", new string[] { "employeeID", "displayName", "sAMAccountName", "department", "st" } } 
            };
            
            rules.ChangeFlowEnd += new ChangeHandler(rules_ChangeFlowEnd);
        }

        void rules_ChangeFlowEnd(ChangeSet changes)
        {
            if (changes.Changes.ContainsKey("sAMAccountName") || changes.Changes.ContainsKey("displayName"))
            {
                var template = Mail.GetTemplate("NameChange");
                var to = Config.App.AppSettings["adminTo"];
                Mail.Queue(to, template.Subject, template.Message, changes.Hologram, changes.Changes);
            }
        }

        public virtual void Terminate()
        {
            // Triggers the final ExportFlowEnd event
            rules.Dispose();
        }

        # region Unused IMASynchronization Methods

        public virtual void Initialize() { }

        public virtual bool ShouldProjectToMV(CSEntry csentry, out string MVObjectType)
        {
            throw new EntryPointNotImplementedException();
        }

        public virtual DeprovisionAction Deprovision(CSEntry csentry)
        {
            throw new EntryPointNotImplementedException();
        }

        public virtual bool FilterForDisconnection(CSEntry csentry)
        {
            throw new EntryPointNotImplementedException();
        }

        public virtual void MapAttributesForJoin(string FlowRuleName, CSEntry csentry, ref ValueCollection values)
        {
            throw new EntryPointNotImplementedException();
        }

        public virtual bool ResolveJoinSearch(string joinCriteriaName, CSEntry csentry, MVEntry[] rgmventry, out int imventry, ref string MVObjectType)
        {
            throw new EntryPointNotImplementedException();
        }

        # endregion

        /// <summary>
        /// To process Rules for import or export, simply invoke the Process method on the Parser object.
        /// </summary>
        /// <param name="FlowRuleName">
        /// The FlowRuleName in this case contains the Rule definition. 
        /// e.g. "IIF(IsNullOrEmpty(mventry.TargetAttribute), Trim(SourceAttribute), Skip())->TargetAttribute"
        /// </param>
        /// <param name="csentry">
        /// CS/MVEntry gets implicitly converted to an EntryWrapper used by the rule engine.
        /// The EntryWrapper unifies the CSEntry and MVEntry objects with a common class so they can 
        /// be used in rules interchangeably.
        /// </param>
        /// <param name="mventry">See description for csentry</param>
        public virtual void MapAttributesForImport(string FlowRuleName, CSEntry csentry, MVEntry mventry)
        {
            try
            {
                rules.Run(FlowRuleName, csentry, mventry);
            }
            catch (Exception e)
            {
                LogWriter.writeEntry("FIM", "Import Error", "ExampleExtension", LogType.Error, 
                    "FlowRule: {0}\r\nError: {1}\r\nTrace: {2}", FlowRuleName, e.Message, e.StackTrace);
            }
        }

        public virtual void MapAttributesForExport(string FlowRuleName, MVEntry mventry, CSEntry csentry)
        {
            try
            {
                rules.Run(FlowRuleName, mventry, csentry);
            }
            catch (Exception e)
            {
                LogWriter.writeEntry("FIM", "Export Error", "ExampleExtension", LogType.Error,
                    "FlowRule: {0}\r\nError: {1}\r\nTrace: {2}", FlowRuleName, e.Message, e.StackTrace);
            }
        }
    }
}
