﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.MetadirectoryServices;

namespace SoftwareIDM.Uplift
{
    /// <summary>
    /// This class provides a lazy binding representation of a value for use with the rule parser.
    /// An AttributeValue can be constructed either by giving it a literal value such as a string or an int,
    /// or by giving it a Rule, which will be evaluated when one of the typed Value accessors is invoked.
    /// </summary>
    public class AttributeValue : IEquatable<AttributeValue>, IComparable<AttributeValue>
    {
        public AttributeValue() { }

        /// <summary>
        /// Construct with a value generating Rule
        /// </summary>
        /// <param name="rule">Rule to be evaluated</param>
        public AttributeValue(Rule rule)
        {
            if (rule == null)
            {
                this.Value = null;
            }

            this.rule = rule;
        }

        /// <summary>
        /// Construct with a literal value, which is type converted as appropriate to one of the
        /// FIM supported primitive values. This constructor does not support multi-valued objects.
        /// </summary>
        /// <param name="val">Value to represent</param>
        public AttributeValue(object val)
        {
            _Value = val;
            if (val == null)
            {
                _DataType = AttributeType.String;
                return;
            }
            else if (val is bool)
            {
                _DataType = AttributeType.Boolean;
            }
            else if (val is long)
            {
                _DataType = AttributeType.Integer;
            }
            else if (val is int)
            {
                _DataType = AttributeType.Integer;
            }
            else if (val is ReferenceValue)
            {
                _DataType = AttributeType.Reference;
            }
            else if (val is byte[])
            {
                _DataType = AttributeType.Binary;
            }
            else
            {
                _DataType = AttributeType.String;
            }
        }

        public ConnectedMA MA { get; set; }
        public ManagementAgent ManagementAgent { get; set; }

        public object Coerce(AttributeType target)
        {
            if (IsMultivalued)
            {
                if (_Values.Count > 0)
                {
                    return new AttributeValue(Values[0]).Coerce(target);
                }

                return null;
            }

            switch (target)
            {
                case AttributeType.Binary:
                    if (_Value == null)
                        return null;

                    switch (DataType)
                    {
                        case AttributeType.Binary:
                            return (byte[])_Value;
                        case AttributeType.Reference:
                            return new AttributeValue(_Value.ToString()).Coerce(target);
                        case AttributeType.Boolean:
                            return (bool)_Value ? new byte[] { 1 } : new byte[] { 0 };
                        case AttributeType.Integer:
                            return BitConverter.GetBytes((long)_Value);
                        default:
                            try
                            {
                                return Convert.FromBase64String((string)_Value);
                            }
                            catch { }
                            return UTF8Encoding.GetEncoding(0).GetBytes((string)_Value);
                    }
                case AttributeType.Boolean:
                    if (_Value == null)
                        return false;
                    switch (DataType)
                    {
                        case AttributeType.Binary:
                            return ((byte[])_Value)[0] > 0;
                        case AttributeType.Reference:
                            return new AttributeValue(_Value.ToString()).Coerce(target);
                        case AttributeType.Boolean:
                            return (bool)_Value;
                        case AttributeType.Integer:
                            return (long)_Value > 0;
                        default:
                            return !String.IsNullOrEmpty((string)_Value) && ((string)_Value).ToLower() != "false";
                    }
                case AttributeType.Integer:
                    if (_Value == null)
                        return (long)0;
                    switch (DataType)
                    {
                        case AttributeType.Binary:
                            return BitConverter.ToInt64((byte[])_Value, 0);
                        case AttributeType.Reference:
                            return long.Parse(_Value.ToString());
                        case AttributeType.Boolean:
                            return (long)((bool)_Value ? 1 : 0);
                        case AttributeType.Integer:
                            if (_Value is int)
                                return (long)(int)_Value;
                            if (_Value is long)
                                return _Value;
                            return long.Parse(_Value.ToString());
                        default:
                            return long.Parse((string)_Value);
                    }
                case AttributeType.Reference:
                    if (_Value == null)
                        return null;

                    switch (DataType)
                    {
                        case AttributeType.Reference:
                            return (ReferenceValue)_Value;
                        default:
                            if (MA != null)
                            {
                                return MA.EscapeDNComponent(_Value.ToString());
                            }
                            else if (ManagementAgent != null)
                            {
                                return ManagementAgent.EscapeDNComponent(_Value.ToString());
                            }

                            throw new InvalidOperationException("Cannot convert to reference value without MA handle");
                    }
                default: // String
                    if (_Value == null)
                        return null;
                    switch (DataType)
                    {
                        case AttributeType.Binary:
                            return Convert.ToBase64String((byte[])_Value);
                        case AttributeType.Boolean:
                            return ((bool)_Value).ToString().ToLower();
                        case AttributeType.Reference:
                            return _Value.ToString();
                        case AttributeType.Integer:
                            return ((long)_Value).ToString();
                        default:
                            return (string)_Value;
                    }
            }
        }

        #region build multivalue

        public static AttributeValue MultiValue(IEnumerable<string> values, MultiBehavior behavior)
        {
            var ret = new AttributeValue();
            ret.IsMultivalued = true;
            ret.MultivalueBehavior = behavior;
            ret.Values = Utils.ValueCollection(values.ToArray());
            return ret;
        }

        public static AttributeValue MultiValue(IEnumerable<ReferenceValue> values, MultiBehavior behavior)
        {
            var ret = MultiValue(new string[] { }, behavior);
            foreach (var val in values)
            {
                ret.Values.Add(val);
            }

            return ret;
        }

        public static AttributeValue MultiValue(IEnumerable<byte[]> values, MultiBehavior behavior)
        {
            var ret = MultiValue(new string[] { }, behavior);
            foreach (var val in values)
            {
                ret.Values.Add(val);
            }

            return ret;
        }

        public static AttributeValue MultiValue(IEnumerable<long> values, MultiBehavior behavior)
        {
            var ret = MultiValue(new string[] { }, behavior);
            foreach (var val in values)
            {
                ret.Values.Add(val);
            }
            return ret;
        }

        #endregion

        /// <summary>
        /// Type coercing equals comparison e.g. AttributeValue("1.0").Equals(AttributeValue(1.0) returns true.
        /// </summary>
        /// <param name="compare">AttributeValue to compare to</param>
        /// <returns>Equality result</returns>
        public bool Equals(AttributeValue compare)
        {
            if (IsMultivalued)
            {
                if (Values == null || compare.Values == null)
                {
                    return Values == null && compare.Values == null;
                }

                if (compare.Values.Count != Values.Count)
                {
                    return false;
                }

                var c = compare.Values.ToStringArray();
                foreach (string s in Values.ToStringArray())
                {
                    if (!c.Contains(s))
                    {
                        return false;
                    }
                }

                return true;
            }
            else
            {
                if (Value == null)
                    return compare.Value == null;
                else if (compare.Value == null)
                    return false;
            }

            switch (DataType)
            {
                case AttributeType.Binary:
                    return Convert.ToBase64String(BinaryValue) == Convert.ToBase64String(compare.BinaryValue);
                case AttributeType.Boolean:
                    return BooleanValue == compare.BooleanValue;
                case AttributeType.Integer:
                    return IntegerValue == compare.IntegerValue;
                case AttributeType.Reference:
                    if (Value == null || compare.Value == null)
                    {
                        return Value == null && compare.Value == null;
                    }

                    return Value.ToString() == compare.Value.ToString();
                default:
                    return StringValue == compare.StringValue;
            }
        }

        private int CompareBin(byte[] me, byte[] other)
        {
            if (me.Length > other.Length)
            {
                var hold = new byte[me.Length];
                other.CopyTo(hold, me.Length - other.Length);
                other = hold;
            }
            else if (other.Length > me.Length)
            {
                var hold = new byte[other.Length];
                me.CopyTo(hold, other.Length - me.Length);
                me = hold;
            }

            var ret = 0;
            for (int i = 0; i < me.Length; i++)
            {
                if (me[i] != other[i])
                {
                    ret = me[i].CompareTo(other[i]);
                    break;
                }
            }

            return ret;
        }

        public int CompareTo(AttributeValue compare)
        {

            if (IsMultivalued)
            {
                throw new FormatException("Cannot perform comparison on multivalued attribute");
            }

            if (Value == null)
            {
                if (compare.Value == null)
                    return 0;
                else
                    return int.MinValue;
            }

            if(compare.Value == null){
                return int.MaxValue;
            }

            switch (DataType)
            {
                case AttributeType.Binary:
                    return CompareBin(BinaryValue, compare.BinaryValue);
                case AttributeType.Boolean:
                    return BooleanValue.CompareTo(compare.BooleanValue);
                case AttributeType.Integer:
                    return IntegerValue.CompareTo(compare.IntegerValue);
                case AttributeType.Reference:
                    return ReferenceValue.ToString().CompareTo(compare.ReferenceValue.ToString());
                default:
                    try
                    {
                        return DateTime.Parse(StringValue).CompareTo(DateTime.Parse(compare.StringValue));
                    }
                    catch
                    {
                        return StringValue.CompareTo(compare.StringValue);
                    }
            }
        }

        private void LateBind()
        {
            if (!_DataType.HasValue)
            {
                if (rule != null)
                {
                    var val = rule.LateBind();
                    if (val.SkipFlow)
                    {
                        SkipFlow = val.SkipFlow;
                    }

                    if (val.LastContributionTime.HasValue)
                    {
                        LastContributionTime = val.LastContributionTime;
                        LastContributingMA = val.LastContributingMA;
                    }

                    if (val.IsMultivalued)
                    {
                        IsMultivalued = true;
                        Values = val.Values;
                        MultivalueBehavior = val.MultivalueBehavior;
                        _DataType = AttributeType.String;
                    }
                    else
                    {
                        _Value = val.Value;
                        _DataType = val.DataType;
                    }
                }
                else
                {
                    _DataType = AttributeType.String;
                }
            }
        }

        AttributeType? _DataType = null;
        /// <summary>
        /// Enumeration representing the current Type that's being stored in the attribute value.
        /// </summary>
        public AttributeType DataType
        {
            get
            {
                LateBind();
                return _DataType.Value;
            }
        }

        object _Value;
        /// <summary>
        /// Return the underlying value as an object without type coercion.
        /// </summary>
        public object Value
        {
            get
            {
                LateBind();
                return _Value;
            }
            set
            {
                _Value = value;
            }
        }

        ValueCollection _Values = null;
        /// <summary>
        /// Multi-valued collection, only supports collections of strings.
        /// </summary>
        public ValueCollection Values
        {
            get
            {
                LateBind();
                return _Values;
            }
            set
            {
                IsMultivalued = true;
                _DataType = AttributeType.String;
                _Values = value;
            }
        }

        bool _SkipFlow;
        /// <summary>
        /// Special case value. If set, ultimately the Rule engine will skip assigning the value.
        /// It's useful in the case of rules that only execute once to populate a value.
        /// </summary>
        public bool SkipFlow
        {
            get
            {
                LateBind();
                return _SkipFlow;
            }
            set
            {
                _SkipFlow = value;
            }
        }

        Rule rule = null;

        public string StringValue
        {
            get
            {
                LateBind();
                return (string)Coerce(AttributeType.String);
            }
            set
            {
                _DataType = AttributeType.String;
                _Value = value;
            }
        }

        public bool BooleanValue
        {
            get
            {
                LateBind();
                return (bool)Coerce(AttributeType.Boolean);
            }
            set
            {
                _DataType = AttributeType.Boolean;
                _Value = value;
            }
        }

        public byte[] BinaryValue
        {
            get
            {
                LateBind();
                return (byte[])Coerce(AttributeType.Binary);
            }
            set
            {
                _DataType = AttributeType.Binary;
                _Value = value;
            }
        }

        public long IntegerValue
        {
            get
            {
                LateBind();
                return (long)Coerce(AttributeType.Integer);
            }
            set
            {
                _DataType = AttributeType.Integer;
                _Value = value;
            }
        }

        public ReferenceValue ReferenceValue
        {
            get
            {
                LateBind();
                return (ReferenceValue)Coerce(AttributeType.Reference);
            }
            set
            {
                _DataType = AttributeType.Reference;
                _Value = value;
            }
        }
        public bool IsMultivalued = false;
        public MultiBehavior MultivalueBehavior = MultiBehavior.Append;
        /// <summary>
        /// Shortcut to return the string value with the Trim method called on it.
        /// This is a useful shortcut, because it automatically returns "" for a Null string.
        /// </summary>
        /// <param name="trim">chars to trim, passed directly to String.Trim()</param>
        /// <returns>Trimmed String</returns>
        public string TrimmedString(params char[] trim)
        {
            if (StringValue == null || String.IsNullOrEmpty(StringValue.Trim(trim)))
            {
                return "";
            }
            return StringValue;
        }

        public override string ToString()
        {
            if (IsMultivalued)
            {
                var ret = new StringBuilder();
                foreach (var val in Values)
                {
                    ret.Append(val.ToString() + "\r\n");
                }

                return ret.ToString();
            }

            switch (DataType)
            {
                case AttributeType.Reference:
                    return ReferenceValue == null ? "" : ReferenceValue.ToString();
                case AttributeType.Integer:
                    return IntegerValue.ToString();
                case AttributeType.Boolean:
                    return BooleanValue.ToString();
                case AttributeType.Binary:
                    return BinaryValue == null ? "" : Convert.ToBase64String(BinaryValue);
                default:
                    return StringValue == null ? "" : StringValue;
            }
        }

        public override int GetHashCode()
        {
            return _Value == null ? base.GetHashCode() : _Value.GetHashCode();
        }

        ManagementAgent _LastContributingMA;
        public ManagementAgent LastContributingMA
        {
            get
            {
                LateBind();
                return _LastContributingMA;
            }
            set { _LastContributingMA = value; }
        }

        DateTime? _LastContributionTime;
        public DateTime? LastContributionTime
        {
            get
            {
                LateBind();
                return _LastContributionTime;
            }
            set { _LastContributionTime = value; }
        }
    }

    public enum MultiBehavior
    {
        Append,
        Replace
    }
}
