﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.MetadirectoryServices;

namespace SoftwareIDM.Uplift
{
    /// <summary>
    /// Common functions based on capabilities of FIM Portal.
    /// Includes signatures of all FIM Portal functions and a few others that may be useful.
    /// 
    /// For a function to be used in a Rule, it must return an AttributeValue and accept only
    /// AttributeValues as arguments. 
    /// It should also be a static function, and should not cache values that may vary from user to user.
    /// 
    /// For additional information and examples for many of these functions, 
    /// refer to FIM 2010 function reference.
    /// http://technet.microsoft.com/en-us/library/ff800820%28WS.10%29.aspx
    /// </summary>
    public static class CommonFunctions
    {
        /// <summary>
        /// Invoke bitwise & operator on two integers.
        /// </summary>
        public static AttributeValue BitAnd(AttributeValue mask, AttributeValue flag)
        {
            return new AttributeValue(flag.IntegerValue & mask.IntegerValue);
        }

        /// <summary>
        /// Invoke bitwise | operator on two integers
        /// </summary>
        public static AttributeValue BitOr(AttributeValue mask, AttributeValue flag)
        {
            return new AttributeValue(flag.IntegerValue | mask.IntegerValue);
        }

        /// <summary>
        /// Invoke  ~ on an int
        /// </summary>
        public static AttributeValue BitNot(AttributeValue flag)
        {
            return new AttributeValue(~flag.IntegerValue);
        }

        /// <summary>
        /// Constant for a Windows style newline.
        /// </summary>
        public static AttributeValue CRLF()
        {
            return new AttributeValue { StringValue = "\r\n" };
        }

        /// <summary>
        /// Formats an AttributeValue parseable by the DateTime class using DateTime.ToString(format)
        /// formatting rules.
        /// </summary>
        public static AttributeValue DateTimeFormat(AttributeValue dateTimeString, AttributeValue format)
        {
            try
            {
                return new AttributeValue(DateTime.Parse(dateTimeString.StringValue).ToString(format.StringValue));
            }
            catch (FormatException)
            {
                return new AttributeValue();
            }
        }

        public static AttributeValue ToFileTime(AttributeValue dateTimeString)
        {
            try
            {
                return new AttributeValue(DateTime.Parse(dateTimeString.StringValue).ToFileTime());
            }
            catch (FormatException)
            {
                return new AttributeValue();
            }
        }

        /// <summary>
        /// Formats an AttributeValue parseable by the DateTime class as a UTC time suited for the FIM Portal
        /// </summary>
        public static AttributeValue ToPortalTime(AttributeValue dateTimeString)
        {
            if (!String.IsNullOrEmpty(dateTimeString.StringValue))
            {
                return new AttributeValue(DateTime.Parse(dateTimeString.StringValue).ToUniversalTime().ToString("yyyy-MM-ddTHH:mm:ss.000"));
            }

            return CommonFunctions.Null();
        }

        public static AttributeValue FromFileTime(AttributeValue time)
        {
            return new AttributeValue(DateTime.FromFileTime(time.IntegerValue));
        }

        public static AttributeValue Now()
        {
            return new AttributeValue(DateTime.Now);
        }

        public static AttributeValue MaxInt()
        {
            return new AttributeValue(long.MaxValue);
        }

        public static AttributeValue MinInt()
        {
            return new AttributeValue(long.MinValue);
        }

        public static AttributeValue Multi(AttributeValue multi, AttributeValue mode)
        {
            var ret = new AttributeValue();
            ret.IsMultivalued = true;
            ret.SkipFlow = multi.SkipFlow;
            if (multi.IsMultivalued)
            {
                ret.Values = multi.Values;
            }
            else if (multi.StringValue != null)
            {
                ret.Values = Utils.ValueCollection(multi.StringValue);
            }

            if (mode != null && mode.StringValue.ToLower().Trim() == "append")
            {
                ret.MultivalueBehavior = MultiBehavior.Append;
            }
            else
            {
                ret.MultivalueBehavior = MultiBehavior.Replace;
            }

            return ret;
        }

        /// <summary>
        /// Invokes Microsoft.MetadirectoryServices.Utils.ConvertSidToString on a byte[]
        /// </summary>
        public static AttributeValue ConvertSidToString(AttributeValue sid)
        {
            return new AttributeValue(Microsoft.MetadirectoryServices.Utils.ConvertSidToString(sid.BinaryValue));
        }

        /// <summary>
        /// Function to perform a conditional branch. Due to lazy binding of AttributeValues, 
        /// if Rule based attributes are passed to IIF, only the Rule on the followed branch gets executed.
        /// This allows rules with side-effects to be safely used with the IIF function.
        /// </summary>
        /// <param name="condition">condition must evaluate or coerce to a BooleanValue</param>
        public static AttributeValue IIF(AttributeValue condition, AttributeValue valueTrue, AttributeValue valueFalse)
        {
            return condition.BooleanValue ? valueTrue : valueFalse;
        }

        /// <summary>
        /// Invokes String.Join with params[]
        /// </summary>
        public static AttributeValue Join(AttributeValue separator, params AttributeValue[] values)
        {
            return new AttributeValue(String.Join(separator.StringValue,
                    (from v in values
                     where !String.IsNullOrEmpty(v.StringValue)
                     select v.StringValue).ToArray()));
        }

        /// <summary>
        /// Returns first element in values that isn't null. Should be used with caution on Boolean and Integer fields,
        /// since it may skip a Boolean that is false, or an Integer that is 0.
        /// </summary>
        public static AttributeValue FirstNotNull(params AttributeValue[] values)
        {
            foreach (var val in values)
            {
                if (val.Value != null)
                {
                    return val;
                }
            }
            return CommonFunctions.Null();
        }

        /// <summary>
        /// Returns a SafeSubstring of numchars from the Left side of the string.
        /// </summary>
        public static AttributeValue Left(AttributeValue str, AttributeValue numchars)
        {
            return new AttributeValue(str.StringValue.SafeSubstring(0, (int)numchars.IntegerValue));
        }

        /// <summary>
        /// Returns a SafeSubstring of numchars from the Right side of the string.
        /// </summary>
        public static AttributeValue Right(AttributeValue str, AttributeValue numchars)
        {
            return new AttributeValue(str.StringValue.SafeSubstring(
                    str.StringValue.Length - (int)numchars.IntegerValue,
                    (int)numchars.IntegerValue));
        }

        /// <summary>
        /// Invokes String.PadLeft(numchars, padCharacter).
        /// Technically padCharacter must be a StringValued AttributeValue, and only index [0] is used.
        /// </summary>
        public static AttributeValue LeftPad(AttributeValue str, AttributeValue numchars, AttributeValue padCharacter)
        {
            return String.IsNullOrEmpty(str.StringValue) ?
                str :
                new AttributeValue(str.StringValue.PadLeft((int)numchars.IntegerValue, padCharacter.StringValue[0]));
        }

        /// <summary>
        /// Invokes String.PadRight(numchars, padCharacter)
        /// </summary>
        public static AttributeValue RightPad(AttributeValue str, AttributeValue numchars, AttributeValue padCharacter)
        {
            return String.IsNullOrEmpty(str.StringValue) ?
                str :
                new AttributeValue(str.StringValue.PadRight((int)numchars.IntegerValue, padCharacter.StringValue[0]));
        }

        /// <summary>
        /// Invokes str.Replace(oldValue, newValue)
        /// </summary>
        public static AttributeValue ReplaceString(AttributeValue str, AttributeValue oldValue, AttributeValue newValue)
        {
            return String.IsNullOrEmpty(str.StringValue) ?
                str :
                new AttributeValue(str.StringValue.Replace(oldValue.StringValue, newValue.StringValue));
        }

        /// <summary>
        /// SafeSubstring is a helper proxy for String.Substring that performs bounds checking to avoid
        /// index errors. The returned substring may be "", or may be shorter than the length parameter.
        /// SafeSubstring is cannot be used directly by Rules. See Mid.
        /// </summary>
        public static string SafeSubstring(this String str, int start, int length)
        {
            if (str == null || start < 0 || start > str.Length)
            {
                return "";
            }

            int max = str.Length - start;
            if (length <= 0)
            {
                length = max;
            }

            if (max < length)
            {
                length = max;
            }
            return str.Substring(start, length);
        }

        /// <summary>
        /// Mid is a proxy for SafeSubstring that may be used in Rule definitions.
        /// </summary>
        public static AttributeValue Mid(AttributeValue str, AttributeValue pos, AttributeValue numchars)
        {
            if (numchars == null)
            {
                numchars = new AttributeValue(0);
            }

            return new AttributeValue(str.StringValue.SafeSubstring((int)pos.IntegerValue, (int)numchars.IntegerValue));
        }

        /// <summary>
        /// Proxy for String.TrimStart()
        /// </summary>
        public static AttributeValue LTrim(AttributeValue str)
        {
            return String.IsNullOrEmpty(str.StringValue) ?
                str : new AttributeValue(str.StringValue.TrimStart());
        }

        /// <summary>
        /// Proxy for String.TrimEnd()
        /// </summary>
        public static AttributeValue RTrim(AttributeValue str)
        {
            return String.IsNullOrEmpty(str.StringValue) ?
                str : new AttributeValue(str.StringValue.TrimEnd());
        }

        /// <summary>
        /// Proxy for String.Trim(), handles null strings.
        /// </summary>
        public static AttributeValue Trim(AttributeValue str)
        {
            return String.IsNullOrEmpty(str.StringValue) ?
                str : new AttributeValue(str.StringValue.Trim());
        }

        /// <summary>
        /// Converts a string to Title Case e.g. "john DOE" -> "John Doe"
        /// </summary>
        public static AttributeValue ProperCase(AttributeValue str)
        {
            if (!String.IsNullOrEmpty(str.StringValue))
            {
                var ti = System.Threading.Thread.CurrentThread.CurrentCulture.TextInfo;
                return new AttributeValue(ti.ToTitleCase(str.StringValue.ToLower()));
            }

            return str;
        }

        /// <summary>
        /// Invokes String.ToLower
        /// </summary>
        public static AttributeValue LowerCase(AttributeValue str)
        {
            return String.IsNullOrEmpty(str.StringValue) ?
                str : new AttributeValue(str.StringValue.ToLower());
        }

        /// <summary>
        /// Invokes String.ToUpper
        /// </summary>
        public static AttributeValue UpperCase(AttributeValue str)
        {
            return String.IsNullOrEmpty(str.StringValue) ?
                str : new AttributeValue(str.StringValue.ToUpper());
        }

        /// <summary>
        /// Throw an UnexpectedDataException with the given message.
        /// </summary>
        public static AttributeValue UnexpectedDataException(AttributeValue message, params AttributeValue[] args)
        {
            throw new UnexpectedDataException(String.Format(message.StringValue, args));
        }

        /// <summary>
        /// Generates a random number
        /// </summary>
        public static AttributeValue RandomNum(AttributeValue start, AttributeValue end)
        {
            return new AttributeValue(new Random().Next((int)start.IntegerValue, (int)end.IntegerValue));
        }

        /// <summary>
        /// Invokes String.IsNullOrEmpty. Coerces type as needed.
        /// </summary>
        public static AttributeValue IsNullOrEmpty(AttributeValue val)
        {
            bool ret;
            switch (val.DataType)
            {
                case AttributeType.Binary:
                    ret = val.BinaryValue == null || val.BinaryValue.Length == 0;
                    break;
                case AttributeType.Boolean:
                    ret = !val.BooleanValue;
                    break;
                case AttributeType.Integer:
                    ret = val.IntegerValue == 0;
                    break;
                case AttributeType.Reference:
                    ret = val.ReferenceValue == null || String.IsNullOrEmpty(val.ReferenceValue.ToString());
                    break;
                default:
                    ret = String.IsNullOrEmpty(val.StringValue);
                    break;
            }
            return new AttributeValue(ret);
        }

        /// <summary>
        /// Combines params[] AttributeValue.BooleanValues with boolean && operator.
        /// Performs short-circuit evaluation with lazy binding Rule based attributes.
        /// </summary>
        public static AttributeValue And(params AttributeValue[] val)
        {
            var ret = true;
            foreach (var r in val)
            {
                ret = ret && r.BooleanValue;
                if (!ret)
                    break;
            }
            return new AttributeValue(ret);
        }

        /// <summary>
        /// Combines params[] AttributeValue.BooleanValues with boolean || operator.
        /// Performs short-circuit evaluation with lazy binding Rule based attributes.
        /// </summary>
        public static AttributeValue Or(params AttributeValue[] val)
        {
            var ret = false;
            foreach (var r in val)
            {
                ret = ret || r.BooleanValue;
            }
            return new AttributeValue(ret);
        }

        /// <summary>
        /// Negates a boolean AttributeValue
        /// </summary>
        public static AttributeValue Not(AttributeValue val)
        {
            return new AttributeValue(!val.BooleanValue);
        }

        /// <summary>
        /// Invokes AttributeValue.Equals(compare)
        /// </summary>
        public static AttributeValue Eq(AttributeValue val, AttributeValue compare)
        {
            return new AttributeValue(val.Equals(compare));
        }

        /// <summary>
        /// Returns an AttributeValue object representing the Null value
        /// (or 0, false, [] when type appropriate).
        /// </summary>
        public static AttributeValue Null()
        {
            return new AttributeValue();
        }

        /// <summary>
        /// Proxy for String.EndsWith
        /// </summary>
        public static AttributeValue EndsWith(AttributeValue val, AttributeValue compare)
        {
            return val.StringValue == null ?
                new AttributeValue(false) :
                new AttributeValue(val.StringValue.EndsWith(compare.StringValue));
        }

        /// <summary>
        /// Proxy for String.StartsWith
        /// </summary>
        public static AttributeValue StartsWith(AttributeValue val, AttributeValue compare)
        {
            return val.StringValue == null ?
                new AttributeValue(false) :
                new AttributeValue(val.StringValue.StartsWith(compare.StringValue));
        }

        /// <summary>
        /// Proxy for String.Contains
        /// </summary>
        /// <param name="val"></param>
        /// <param name="compare"></param>
        /// <returns></returns>
        public static AttributeValue Contains(AttributeValue val, AttributeValue compare)
        {
            if (val.Value == null)
            {
                return new AttributeValue(false);
            }
            return new AttributeValue(val.StringValue.Contains(compare.StringValue));
        }

        /// <summary>
        /// Constructs an AttributeValue with SkipFlow set to true.
        /// </summary>
        public static AttributeValue Skip()
        {
            var ret = new AttributeValue();
            ret.SkipFlow = true;
            return ret;
        }

        /// <summary>
        /// Convert scope and type into GroupType
        /// </summary>
        public static AttributeValue ADGroupType(AttributeValue scope, AttributeValue type)
        {
            switch (type.StringValue)
            {
                case "Distribution":
                    switch (scope.StringValue)
                    {
                        case "DomainLocal":
                            return new AttributeValue(Config.ADConst.DOMAINLOCAL_DIST_GRP);
                        case "Global":
                            return new AttributeValue(Config.ADConst.GLOBAL_DIST_GRP);
                        default: // "Universal":
                            return new AttributeValue(Config.ADConst.UNIVERSAL_DIST_GRP);
                    }
                default: // Security, MailEnabledSecurity
                    switch (scope.StringValue)
                    {
                        case "DomainLocal":
                            return new AttributeValue(Config.ADConst.DOMAINLOCAL_SEC_GRP);
                        case "Global":
                            return new AttributeValue(Config.ADConst.GLOBAL_SEC_GRP);
                        default: // "Universal":
                            return new AttributeValue(Config.ADConst.UNIVERSAL_DIST_GRP);
                    }
            }
        }

        /// <summary>
        /// Gets OU component of an AD DN
        /// </summary>
        public static AttributeValue ParseOU(AttributeValue dn)
        {
            if (dn != null && !String.IsNullOrEmpty(dn.StringValue))
            {
                return new AttributeValue(Config.ADConst.GetContainer(dn.StringValue));
            }

            return CommonFunctions.Null();
        }

        /// <summary>
        /// Gets CN component of an AD DN
        /// </summary>
        public static AttributeValue ParseCN(AttributeValue dn)
        {
            if (dn != null && !String.IsNullOrEmpty(dn.StringValue))
            {
                return new AttributeValue(Config.ADConst.GetCN(dn.StringValue));
            }

            return CommonFunctions.Null();
        }

        /// <summary>
        /// returns scope component of group type as DomainLocal, Global, Universal
        /// </summary>
        public static AttributeValue GroupScope(AttributeValue groupType)
        {
            var gt = groupType.IntegerValue;
            return new AttributeValue((gt & 2) == 2 ? "Global" : ((gt & 4) == 4 ? "DomainLocal" : "Universal"));
        }

        /// <summary>
        /// returns type component of group type as Distribution, Security
        /// </summary>
        public static AttributeValue GroupType(AttributeValue groupType)
        {
            var gt = groupType.IntegerValue;
            return new AttributeValue(gt > 0 ? "Distribution" : "Security");
        }
    }
}
