﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;
using System.Xml.Linq;
using System.Data;
using Microsoft.MetadirectoryServices;

namespace SoftwareIDM.Uplift
{

    public interface IConfig
    {
        void Set(XElement xml);
        string Name { get; }
    }

    public static class Config
    {
        static XElement _xml;
        static Dictionary<string, IConfig> _data;
        public static Dictionary<string, IConfig> Data
        {
            get
            {
                if (_data == null)
                {
                    _data = new Dictionary<string, IConfig>();
                }

                return _data;
            }
        }

        public static T Add<T>() where T : IConfig
        {
            var add = (T)typeof(T).GetConstructor(new Type[] { }).Invoke(new object[] { });
            add.Set(Xml);
            Data.Add(add.Name, add);
            return add;
        }

        public static T Get<T>() where T : IConfig
        {
            var add = (T)typeof(T).GetConstructor(new Type[] { }).Invoke(new object[] { });
            if (!Data.ContainsKey(add.Name))
            {
                add.Set(Xml);
                Data.Add(add.Name, add);
            }

            return (T)Data[add.Name];
        }

        public static XElement Xml
        {
            get
            {
                if (_xml == null)
                {
                    _xml = XElement.Load(ExeConfigPath == null ? Utils.ExtensionsDirectory + @"\App.config" : ExeConfigPath);
                }

                return _xml;
            }
        }

        static Configuration _app;
        public static Configuration App
        {
            get
            {
                if (_app == null)
                {
                    Add<Configuration>();
                    _app = (Configuration)Data["Application"];
                }

                return _app;
            }
        }

        static Map _map;
        public static Map Map
        {
            get
            {
                if (_map == null)
                {
                    _map = new Map();
                }
                return _map;
            }
        }

        static AD _ADConst;
        public static AD ADConst
        {
            get
            {
                if (_ADConst == null)
                {
                    _ADConst = new AD();
                }
                return _ADConst;
            }
        }

        static string _exeConfigPath;
        public static string ExeConfigPath
        {
            get { return _exeConfigPath; }
            set
            {
                if (!value.Contains('\\'))
                {
                    _exeConfigPath = Utils.ExtensionsDirectory + "\\" + value;
                }
                else
                {
                    _exeConfigPath = value;
                }
                var access = App;
            }
        }
    }

    public class ConnectionConfig
    {
        public string Name { get; set; }
        public string ConnectionString { get; set; }
    }

    public class Configuration : IConfig
    {
        public string Name { get { return "Application"; } }

        public XElement xml { get; set; }
        public Dictionary<string, Log> Logs { get; set; }
        public MailConfig Mail { get; set; }
        public Dictionary<string, string> AppSettings { get; set; }
        public Dictionary<string, ConnectionConfig> ConnectionStrings { get; set; }

        Dictionary<string, Object> Sections { get; set; }

        public void Set(XElement xml)
        {
            this.xml = xml;
            Sections = new Dictionary<string, object>();

            if (xml.Elements("Logging").Count() > 0)
            {
                Logs = ((LogConfig)Config.Add<LogConfig>()).Logs;
            }
            else
            {
                Logs = new Dictionary<string, Log>();
            }

            if (xml.Elements("Mail").Count() > 0)
            {
                Mail = (MailConfig)Config.Add<MailConfig>();
            }
            else
            {
                Mail = null;
            }

            AppSettings = new Dictionary<string, string>();
            ConnectionStrings = new Dictionary<string, ConnectionConfig>();
            if (xml.Elements("appSettings").Count() > 0)
            {
                foreach (var app in xml.Element("appSettings").Elements("add"))
                {
                    AppSettings[app.Attribute("key").Value] = app.Attribute("value").Value;
                }
            }

            if (xml.Elements("connectionStrings").Count() > 0)
            {
                foreach (var conn in xml.Element("connectionStrings").Elements("add"))
                {
                    ConnectionStrings[conn.Attribute("name").Value] =
                        new ConnectionConfig { Name = conn.Attribute("name").Value, ConnectionString = conn.Attribute("connectionString").Value };
                }
            }
        }
    }

    public class Map
    {
        Dictionary<string, Dictionary<string, Dictionary<string, string>>> maps = new Dictionary<string, Dictionary<string, Dictionary<string, string>>>();

        Dictionary<string, XElement> xmls = new Dictionary<string, XElement>();

        /// <summary>
        /// Loads xml so that XElement is available to query with Linq, but does not parse into dictionary.
        /// </summary>
        /// <param name="name">Name to reference xml file by.</param>
        /// <param name="path">Location of xml file.</param>
        public void Load(string name, string path)
        {
            var file = XElement.Load(path);
            if (xmls.ContainsKey(name))
            {
                throw new ArgumentException("An xml file named " + name + " has already been loaded.");
            }

            xmls[name] = file;
        }

        /// <summary>
        /// Calls Add(...) with default values of "key" and true for key and xmlIsAttributes.
        /// </summary>
        public void Add(string name, string path, string element)
        {
            Add(name, path, element, "key", true);
        }

        /// <summary>
        /// Adds a new xml mapping file.
        /// </summary>
        /// <param name="name">The name to refer to the file by</param>
        /// <param name="path">The location of the file</param>
        /// <param name="element">The name xml elements in the file that contain the pertinent data</param>
        /// <param name="key">The key field to lookup data by (must be unique)</param>
        /// <param name="xmlIsAttributes">Whether the xml data is stored in xml attibutes on the element, or sub-elements.</param>
        public void Add(string name, string path, string element, string key, bool xmlIsAttributes)
        {
            string val = "";
            var add = new Dictionary<string, Dictionary<string, string>>();
            var map = XElement.Load(path);
            if (xmls.ContainsKey(name))
            {
                throw new ArgumentException("A map file named " + name + " has already been loaded.");
            }

            try
            {
                xmls[name] = map;
                foreach (XElement el in map.Descendants(element))
                {
                    if (xmlIsAttributes)
                    {
                        if (el.Attributes(key).Count() == 0)
                        {
                            throw new ArgumentException("The key attribute " + key + " is not present on the xml row.");
                        }
                        val = el.Attribute(key).Value;
                        add.Add(val, new Dictionary<string, string>());
                        foreach (var attr in el.Attributes())
                        {
                            add[val][attr.Name.ToString()] = attr.Value;
                        }
                    }
                    else
                    {
                        if (el.Elements(key).Count() == 0)
                        {
                            throw new ArgumentException("The key element " + key + " is not present on the xml row.");
                        }
                        val = el.Element(key).Value;
                        add.Add(val, new Dictionary<string, string>());
                        foreach (var attr in el.Elements())
                        {
                            add[val][attr.Name.ToString()] = attr.Value;
                        }
                    }

                    maps[name] = add;
                }
            }
            catch (ArgumentException e)
            {
                if (e.Message.Contains("An item with the same key has already been added"))
                {
                    throw new ArgumentException("A record with the key " + val + " has already been loaded.");
                }

                throw e;
            }
        }

        /// <summary>
        /// Lookup the XElement root of the loaded xml file by name.
        /// </summary>
        /// <param name="name">name of the xml file (see Add)</param>
        /// <returns>XElement representation of xml</returns>
        public XElement Get(string name)
        {
            return xmls[name];
        }

        /// <summary>
        /// Looks up a single value in the xml file.
        /// </summary>
        /// <param name="name">name of the xml file (see Add)</param>
        /// <param name="key">key value to lookup</param>
        /// <param name="field">field from the xml file with the desired value.</param>
        /// <returns></returns>
        public string Get(string name, string key, string field)
        {
            var map = maps[name];
            if (!map.ContainsKey(key) || !map[key].ContainsKey(field))
            {
                return null;
            }

            return map[key][field];
        }
    }

    public class AD
    {
        // userAccountControl constants
        public long SCRIPT = 0x0001;
        public long ACCOUNTDISABLE = 0x0002;
        public long HOMEDIR_REQUIRED = 0x0008;
        public long LOCKOUT = 0x0010;
        public long PASSWD_NOTREQD = 0x0020;
        public long NORMAL_ACCOUNT = 0x0200;
        public long DONT_EXPIRE_PASSWORD = 0x10000;
        public long SMARTCARD_REQUIRED = 0x40000;
        public long TRUSTED_FOR_DELEGATION = 0x80000;
        public long PASSWORD_EXPIRED = 0x800000;

        // groupType constants
        public long GLOBAL_DIST_GRP = 2;
        public long DOMAINLOCAL_DIST_GRP = 4;
        public long UNIVERSAL_DIST_GRP = 8;

        public long GLOBAL_SEC_GRP = -2147483646;
        public long DOMAINLOCAL_SEC_GRP = -2147483644;
        public long UNIVERSAL_SEC_GRP = -2147483640;

        public Regex DNParse = new Regex(@"^(.*?),(?=OU=|DC=|CN=)(.*)$");

        public string GetCN(string dn)
        {
            if (DNParse.IsMatch(dn))
            {
                return DNParse.Match(dn).Groups[1].Value;
            }

            return null;
        }

        public string GetContainer(string dn)
        {
            if (DNParse.IsMatch(dn))
            {
                return DNParse.Match(dn).Groups[2].Value;
            }

            return null;
        }
    }
}
