﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.MetadirectoryServices;

namespace SoftwareIDM.Uplift
{

    public enum EntryWrapperType
    {
        MVEntry,
        CSEntry,
        Dictionary
    }
    /// <summary>
    /// EntryWrapper wraps either a CSEntry or an MVEntry so they may be used in functions
    /// interchangeably.
    /// When it comes to setting and reading values, CS and MV entries behave almost identically, but
    /// they don't inherit from a common parent, so you can't conveniently use the same method signatures 
    /// for both export and import. EntryWrapper resolves that issue.
    /// </summary>
    public class EntryWrapper : IDictionary<string, AttributeValue>
    {
        CSEntry cdata = null;
        MVEntry mdata = null;
        Dictionary<string, AttributeValue> data = null;

        public EntryWrapper(CSEntry csentry)
        {
            cdata = csentry;
            type = EntryWrapperType.CSEntry;
        }

        public EntryWrapper(MVEntry mventry)
        {
            mdata = mventry;
            type = EntryWrapperType.MVEntry;
        }

        public EntryWrapper()
        {
            data = new Dictionary<string, AttributeValue>();
            type = EntryWrapperType.Dictionary;
        }

        public static implicit operator EntryWrapper(CSEntry csentry)
        {
            return new EntryWrapper(csentry);
        }

        public static implicit operator EntryWrapper(MVEntry mventry)
        {
            return new EntryWrapper(mventry);
        }

        /// <summary>
        /// Retrieve the underlying entry object.
        /// </summary>
        public MVEntry mventry
        {
            get { return mdata; }
        }

        /// <summary>
        /// Retrieve the underlying entry object.
        /// </summary>
        public CSEntry csentry
        {
            get { return cdata; }
        }

        EntryWrapperType type;

        /// <summary>
        /// Accessor to check if the wrapped object is a CSEntry or an MVEntry.
        /// </summary>
        public EntryWrapperType Type
        {
            get { return type; }
        }

        /// <summary>
        /// Set a field on the underlying object to a particular AttributeValue.
        /// </summary>
        /// <param name="field">field to set. e.g. mvWrapper.Set("field", CommonFunctions.Null())</param>
        /// <param name="value">AttributeValue to assign. Type Coercion is performed. value.Skip is respected.</param>
        public void Set(string field, AttributeValue value)
        {
            if (value.SkipFlow)
            {
                return;
            }

            switch (type)
            {
                case EntryWrapperType.MVEntry:
                    if (mdata[field].IsMultivalued)
                    {
                        if (value.Values != null)
                        {
                            if (value.MultivalueBehavior == MultiBehavior.Replace ||
                                !mdata[field].IsPresent || mdata[field].Values == null || mdata[field].Values.Count == 0)
                            {
                                mdata[field].Values = value.Values;
                            }
                            else
                            {
                                foreach (var val in value.Values)
                                {
                                    if (!mdata[field].Values.Contains(val))
                                    {
                                        mdata[field].Values.Add(val);
                                    }
                                }
                            }
                        }
                        else
                        {
                            mdata[field].Delete();
                        }
                    }
                    else
                    {
                        object set = Convert(value, mdata[field].DataType);
                        if (set == null)
                        {
                            mdata[field].Delete();
                        }
                        else
                        {
                            switch (mdata[field].DataType)
                            {
                                case AttributeType.Binary:
                                    mdata[field].BinaryValue = (byte[])set;
                                    break;
                                case AttributeType.Boolean:
                                    mdata[field].BooleanValue = (bool)set;
                                    break;
                                case AttributeType.Integer:
                                    mdata[field].IntegerValue = (long)set;
                                    break;
                                case AttributeType.Reference:
                                    mdata[field].ReferenceValue = (ReferenceValue)set;
                                    break;
                                default:
                                    mdata[field].Value = set.ToString();
                                    break;
                            }
                        }
                    }

                    break;
                case EntryWrapperType.CSEntry:
                    //cdata
                    if (field.ToLower() == "dn")
                    {
                        if (Config.ADConst.DNParse.IsMatch(value.StringValue))
                        {
                            cdata.DN = csentry.MA.EscapeDNComponent(Config.ADConst.GetCN(value.StringValue)).Concat(Config.ADConst.GetContainer(value.StringValue));
                        }
                        else
                        {
                            cdata.DN = csentry.MA.EscapeDNComponent(value.StringValue);
                        }

                        return;
                    }

                    if (cdata[field].IsMultivalued)
                    {
                        if (value.Values != null)
                        {
                            if (value.MultivalueBehavior == MultiBehavior.Replace ||
                                !cdata[field].IsPresent || cdata[field].Values == null || cdata[field].Values.Count == 0)
                            {
                                cdata[field].Values = value.Values;
                            }
                            else
                            {
                                foreach (var val in value.Values)
                                {
                                    if (!cdata[field].Values.Contains(val))
                                    {
                                        cdata[field].Values.Add(val);
                                    }
                                }
                            }
                        }
                        else
                        {
                            cdata[field].Delete();
                        }
                    }
                    else
                    {
                        value.ManagementAgent = cdata.MA;
                        object set = Convert(value, cdata[field].DataType);
                        if (set == null)
                        {
                            cdata[field].Delete();
                        }
                        else
                        {

                            switch (cdata[field].DataType)
                            {
                                case AttributeType.Binary:
                                    cdata[field].BinaryValue = (byte[])set;
                                    break;
                                case AttributeType.Boolean:
                                    cdata[field].BooleanValue = (bool)set;
                                    break;
                                case AttributeType.Integer:
                                    cdata[field].IntegerValue = (long)set;
                                    break;
                                case AttributeType.Reference:
                                    cdata[field].ReferenceValue = (ReferenceValue)set;
                                    break;
                                default:
                                    cdata[field].Value = set.ToString();
                                    break;
                            }
                        }
                    }

                    break;
                case EntryWrapperType.Dictionary:
                    data[field.ToLower()] = value;
                    break;
            }
        }

        /// <summary>
        /// Type Coercion utility used by EntryWrapper
        /// </summary>
        /// <param name="value">Value to coerce</param>
        /// <param name="to">AttributeType of receiving field</param>
        /// <param name="csentry">If provided, and needed CSEntry may be used to create ReferenceValues</param>
        /// <returns>Converted object</returns>
        public object Convert(AttributeValue value, AttributeType to)
        {
            if (value == null) { return null; }
            return value.Coerce(to);
        }

        /// <summary>
        /// Returns value of [cs/mv]entry[field].IsPresent.
        /// </summary>
        public bool IsPresent(string field)
        {
            switch (type)
            {
                case EntryWrapperType.CSEntry:
                    return cdata[field].IsPresent;
                case EntryWrapperType.MVEntry:
                    return mdata[field].IsPresent;
                case EntryWrapperType.Dictionary:
                    return data.ContainsKey(field);
                default:
                    return false;
            }
        }

        /// <summary>
        /// Gets the value of the field on the underlying object.
        /// It is safe to call without first invoking IsPresent. If the attribute is not present
        /// An AttributeValue() object representing Null is returned. The Get method
        /// also accesses the special fields DN, RDN, and objectType (not case-sensitive).
        /// </summary>
        public AttributeValue Get(string field)
        {
            var ret = new AttributeValue();

            if (type != EntryWrapperType.Dictionary)
            {
                // Special Fields

                if (type == EntryWrapperType.CSEntry)
                {
                    if (field.ToLower() == "dn")
                    {
                        try
                        {
                            ret.ReferenceValue = cdata.DN;
                        }
                        catch (InvalidOperationException) { } // DN not yet set
                        return ret;
                    }
                    else if (field.ToLower() == "rdn")
                    {
                        ret.StringValue = cdata.RDN;
                        return ret;
                    }
                    else if (field.ToLower() == "objecttype")
                    {
                        ret.StringValue = cdata.ObjectType;
                        return ret;
                    }
                }
                else
                {
                    if (field.ToLower() == "objectid")
                    {
                        ret.StringValue = mdata.ObjectID.ToString();
                        return ret;
                    }
                    else if (field.ToLower() == "objecttype")
                    {
                        ret.StringValue = mdata.ObjectType;
                        return ret;
                    }
                }

                bool isMV = type == EntryWrapperType.MVEntry;

                ret.IsMultivalued = isMV ? mdata[field].IsMultivalued : cdata[field].IsMultivalued;
                if (isMV ? mdata[field].IsPresent : cdata[field].IsPresent)
                {
                    if (isMV)
                    {
                        ret.LastContributionTime = mdata[field].LastContributionTime;
                        ret.LastContributingMA = mdata[field].LastContributingMA;
                    }

                    if (ret.IsMultivalued)
                    {
                        ret.Values = isMV ? mdata[field].Values : cdata[field].Values;
                        ret.MultivalueBehavior = MultiBehavior.Replace;
                        return ret;
                    }


                    switch (isMV ? mdata[field].DataType : cdata[field].DataType)
                    {
                        case AttributeType.Binary:
                            ret.BinaryValue = isMV ? mdata[field].BinaryValue : cdata[field].BinaryValue;
                            break;
                        case AttributeType.Boolean:
                            ret.BooleanValue = isMV ? mdata[field].BooleanValue : cdata[field].BooleanValue;
                            break;
                        case AttributeType.Integer:
                            ret.IntegerValue = isMV ? mdata[field].IntegerValue : cdata[field].IntegerValue;
                            break;
                        case AttributeType.Reference:
                            ret.ReferenceValue = isMV ? mdata[field].ReferenceValue : cdata[field].ReferenceValue;
                            break;
                        default:
                            ret.StringValue = isMV ? mdata[field].Value : cdata[field].Value;
                            break;
                    }
                }
            }
            else
            {
                if (data.ContainsKey(field.ToLower()))
                {
                    ret = data[field.ToLower()];
                }
            }

            return ret;
        }

        public void Add(string key, AttributeValue value)
        {
            throw new NotImplementedException();
        }

        public bool ContainsKey(string key)
        {
            switch (type)
            {
                case EntryWrapperType.MVEntry:
                    return mdata[key].IsPresent;
                case EntryWrapperType.CSEntry:
                    return cdata[key].IsPresent;
                case EntryWrapperType.Dictionary:
                    return data.ContainsKey(key.ToLower());
                default:
                    return false;
            }
        }

        public ICollection<string> Keys
        {
            get
            {
                var ret = new List<string>();
                switch (type)
                {
                    case EntryWrapperType.CSEntry:
                        foreach (var c in mdata)
                        {
                            ret.Add(c);
                        }
                        break;
                    case EntryWrapperType.MVEntry:
                        foreach (var c in cdata)
                        {
                            ret.Add(c);
                        }
                        break;
                    case EntryWrapperType.Dictionary:
                        ret.AddRange(data.Keys);
                        break;
                }

                return ret;
            }
        }

        public bool Remove(string key)
        {
            var ret = false;
            switch (type)
            {
                case EntryWrapperType.CSEntry:
                    if (cdata[key].IsPresent)
                    {
                        cdata[key].Delete();
                        ret = true;
                    }

                    break;
                case EntryWrapperType.MVEntry:
                    if (mdata[key].IsPresent)
                    {
                        mdata[key].Delete();
                        ret = true;
                    }

                    break;
                case EntryWrapperType.Dictionary:
                    ret = data.Remove(key);
                    break;
            }

            return ret;
        }

        public bool TryGetValue(string key, out AttributeValue value)
        {
            throw new NotImplementedException();
        }

        public ICollection<AttributeValue> Values
        {
            get
            {
                var ret = new List<AttributeValue>();
                foreach (var field in Keys)
                {
                    ret.Add(this.Get(field));
                }

                return ret;
            }
        }

        public AttributeValue this[string key]
        {
            get
            {
                return this.Get(key);
            }
            set
            {
                this.Set(key, value);
            }
        }

        public void Add(KeyValuePair<string, AttributeValue> item)
        {
            throw new NotImplementedException();
        }

        public void Clear()
        {
            throw new NotImplementedException();
        }

        public bool Contains(KeyValuePair<string, AttributeValue> item)
        {
            throw new NotImplementedException();
        }

        public void CopyTo(KeyValuePair<string, AttributeValue>[] array, int arrayIndex)
        {
            throw new NotImplementedException();
        }

        public int Count
        {
            get { throw new NotImplementedException(); }
        }

        public bool IsReadOnly
        {
            get { throw new NotImplementedException(); }
        }

        public bool Remove(KeyValuePair<string, AttributeValue> item)
        {
            throw new NotImplementedException();
        }

        public IEnumerator<KeyValuePair<string, AttributeValue>> GetEnumerator()
        {
            foreach (var key in this.Keys)
            {
                yield return new KeyValuePair<string, AttributeValue>(key, this.Get(key));
            }
        }

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }
    }
}
