﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.Data.SqlClient;
using System.Linq;
using System.Text;
using System.Threading;

namespace SoftwareIDM.Uplift
{
    public enum LogType
    {
        Info,
        Alert,
        Error
    }

    /// <summary>
    /// Class to perform one-liner logging to a database table.
    /// </summary>
    public static class LogWriter
    {
        private static LogQueue logger;

        private static void Initialize()
        {
            if (logger == null)
            {
                logger = new LogQueue();
            }
        }

        /// <summary>
        /// Write a LogEntry object asynchronously.
        /// </summary>
        /// <param name="entry">LogEntry to write</param>
        public static void writeEntry(LogEntry entry)
        {
            Initialize();
            logger.Add(entry);
        }

        /// <summary>
        /// Write a Log entry asynchronously.
        /// </summary>
        /// <param name="log">Name of log to write to. e.g. FIM, Sequencer (must be configured in App.config</param>
        /// <param name="_event">Short name for logged event</param>
        /// <param name="caller">Name of application/library writing log entry</param>
        /// <param name="type">Info, Alert, Error</param>
        /// <param name="message">Logged message</param>
        /// <param name="args">Parameters to add to message with String.Format</param>
        public static void writeEntry(string log, string _event, string caller, LogType type, string message, params object[] args)
        {
            Initialize();
            LogEntry entry = new LogEntry();
            entry.Log = log;
            entry.Event = _event;
            entry.Caller = caller;
            entry.Message = String.Format(message, args);
            entry.Type = type;
            logger.Add(entry);
        }

    }// end LogWriter

    /// <summary>
    /// Helper class to make logging asynchronous/non-blocking
    /// </summary>
    class LogQueue
    {
        private Queue<LogEntry> queue = new Queue<LogEntry>();
        private Dictionary<string, bool> tablesVerified = new Dictionary<string, bool>();
        private object lock_ = new object();
        private Thread thread = null;

        public void Start()
        {
            while (queue.Count > 0)
            {
                LogEntry entry = null;
                lock (lock_)
                {
                    entry = queue.Dequeue();
                }
                WriteEntry(entry);
                lock (lock_)
                {
                    if (queue.Count == 0)
                    {
                        Thread.CurrentThread.Abort();
                    }
                }
            }
        }

        public void Add(LogEntry entry)
        {
            lock (lock_)
            {
                queue.Enqueue(entry);
                if (thread == null || !thread.IsAlive)
                {
                    thread = new Thread(new ThreadStart(this.Start));
                    thread.Start();
                }
            }
        }

        private void WriteEntry(LogEntry entry)
        {
            var log = Config.App.Logs[entry.Log];
            using (SqlConnection conn = new SqlConnection(log.connectionString))
            {
                conn.Open();
                if (!tablesVerified.ContainsKey(entry.Log) || !tablesVerified[entry.Log])
                    CreateTable(entry.Log, conn);
                //execute query to log data
                string insert_entry = String.Format(@"
                    INSERT INTO {0} 
                    (event, caller, message, type, _datetime)
                    VALUES
                    (@event, @caller, @message, @type, @_datetime)
                    ", SqlQuote.quote(log.table));
                SqlCommand cmd = new SqlCommand("", conn);
                cmd.Parameters.AddWithValue("@event", entry.Event);
                cmd.Parameters.AddWithValue("@caller", entry.Caller == null ? (object)DBNull.Value : (object)entry.Caller);
                cmd.Parameters.AddWithValue("@message", entry.Message);
                cmd.Parameters.AddWithValue("@type", entry.Type.ToString());
                cmd.Parameters.AddWithValue("@_datetime", DateTime.Now.ToString("s"));

                cmd.CommandText = insert_entry;
                cmd.ExecuteNonQuery();
            }
        }

        private void CreateTable(string type, SqlConnection conn)
        {
            string create_table = @"
                IF NOT EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'{0}') AND type in (N'U'))
                BEGIN
                CREATE TABLE {0}(
	                [event] [varchar](50) NOT NULL,
	                [caller] [varchar](50) NULL,
	                [_datetime] [datetime] NOT NULL,
	                [message] [varchar](max) NOT NULL,
	                [type] [char](5) NULL,
                    [id] [bigint] IDENTITY(1,1) NOT NULL,
                CONSTRAINT {1} PRIMARY KEY CLUSTERED 
                (
	                [id] ASC
                )WITH (PAD_INDEX = OFF, IGNORE_DUP_KEY = OFF) ON [PRIMARY]
                ) ON [PRIMARY]
                END
                ";
            var log = Config.App.Logs[type];
            create_table = String.Format(create_table, SqlQuote.quote(log.table), SqlQuote.quote("PK_" + log.table));

            SqlCommand cmd = new SqlCommand(create_table, conn);
            cmd.ExecuteNonQuery();
            tablesVerified[type] = true;
        }
    }

    public enum LogFields
    {
        Time,
        Event,
        Caller,
        Message,
        Type
    }
    public enum SortOrder
    {
        Ascending,
        Descending
    }

    /// <summary>
    /// Class representation of a log data row.
    /// </summary>
    public class LogEntry : IComparable<LogEntry>
    {
        public LogFields CompareField { get; set; }

        public string Log { get; set; }

        public DateTime Time { get; set; }
        public string Event { get; set; }
        public string Caller { get; set; }
        public string Message { get; set; }
        public LogType Type { get; set; }

        public override string ToString()
        {
            return String.Format("{0} - {1}: {2}", Time.ToShortTimeString(), Event, Message);
        }

        public int CompareTo(LogEntry other)
        {
            switch (CompareField)
            {
                case LogFields.Time:
                    return Time.CompareTo(other.Time);
                case LogFields.Event:
                    return Event.CompareTo(other.Event);
                case LogFields.Caller:
                    if (Caller == null || other.Caller == null)
                        return 0;
                    return Caller.CompareTo(other.Caller);
                case LogFields.Type:
                    return Type.CompareTo(other.Type);
                case LogFields.Message:
                    if (Message == null || other.Message == null)
                        return 0;
                    return Message.CompareTo(other.Message);
                default:
                    return 0;
            }
        }
    }
}
