﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Xml.Linq;
using Microsoft.MetadirectoryServices;

namespace SoftwareIDM.Uplift
{
    public class ProvisionConfig : IConfig
    {
        public Dictionary<string, List<ProvisionRule>> maRules = new Dictionary<string, List<ProvisionRule>>();

        public void Set(XElement xml)
        {
            xml = xml.Element("Provision");
            foreach (var ma in xml.Elements("ma"))
            {
                var name = ma.Attribute("name").Value;
                maRules[name] = new List<ProvisionRule>();
                maRules[name].AddRange(
                    from r in ma.Elements("rule")
                    select new ProvisionRule(r));

            }
        }

        public string Name { get { return "Provision"; } }
    }

    public class ProvisionRule
    {
        public string Name { get; set; }
        public string MVObjectType { get; set; }
        public string CSObjectType { get; set; }
        public string Connector { get; set; }
        public string Condition { get; set; }
        public bool IsExchange { get; set; }
        public List<string> Rules { get; set; }

        public ProvisionRule(XElement xml)
        {
            Name = xml.Attribute("name").Value;
            MVObjectType = xml.Attribute("mvObjectType").Value;
            CSObjectType = xml.Attribute("csObjectType").Value;
            Connector = xml.Attribute("connector").Value;
            Condition = xml.Attributes("condition").Count() > 0 ? xml.Attribute("condition").Value + "->provision" : "";
            IsExchange = xml.Attributes("exchange").Count() > 0 ? (xml.Attribute("exchange").Value.ToLower() == "true") : false;

            Rules = new List<string>();
            foreach (var l in xml.Value.Split('\n'))
            {
                var line = l.Trim();
                if (line.StartsWith("//") || String.IsNullOrEmpty(line))
                {
                    continue;
                }

                Rules.Add(line.Trim());
            }
        }

        readonly Regex reInteger = new Regex(@"(-?\d+)");
        readonly Regex reComparison = new Regex(@"(>|<|>=|<=|==)");
        public bool TestConnector(long connectorCount)
        {
            if (!String.IsNullOrEmpty(Connector))
            {
                var num = reInteger.Match(Connector).Groups[1].Value;
                if (reComparison.IsMatch(Connector))
                {
                    var compare = reComparison.Match(Connector).Groups[1].Value;
                    return ComparisonOperators.Compare(new AttributeValue(connectorCount), new AttributeValue(long.Parse(num)), compare).BooleanValue;
                }
                else
                {
                    return int.Parse(num) == connectorCount;
                }
            }

            return false;
        }
    }

    public class ProvisionChange
    {
        public ProvisionRule Rule { get; set; }
        public CSEntry csentry { get; set; }
        public bool Add { get; set; }
    }

    public class ProvisionEngine : IDisposable
    {
        ProvisionConfig config;
        RuleEngine rules;

        public ProvisionEngine(params Type[] types)
        {
            config = (ProvisionConfig)Config.Add<ProvisionConfig>();
            rules = new RuleEngine(types);
        }

        public List<ProvisionChange> Provision(MVEntry mventry, bool commit)
        {
            var ret = new List<ProvisionChange>();
            foreach (var ma in config.maRules)
            {
                foreach (var rule in ma.Value)
                {
                    var connectedMA = mventry.ConnectedMAs[ma.Key];
                    var connectorCount = (long)connectedMA.Connectors.Count;
                    if (!rule.TestConnector(connectorCount))
                    {
                        continue;
                    }

                    if (mventry.ObjectType.ToLower() != rule.MVObjectType.ToLower())
                    {
                        continue;
                    }

                    if (!String.IsNullOrEmpty(rule.Condition))
                    {
                        var cond = new Rule(rule.Condition);
                        if (!cond.GetValue(mventry, null).BooleanValue)
                        {
                            continue;
                        }
                    }

                    // Execute Rule
                    if (connectorCount == 0)
                    {
                        CSEntry csentry;
                        if (rule.IsExchange && rule.CSObjectType.ToLower() == "user")
                        {
                            var tempWrapper = new EntryWrapper();
                            tempWrapper["ObjectType"] = new AttributeValue(rule.CSObjectType);
                            foreach (var map in rule.Rules)
                            {
                                rules.Run(map, mventry, tempWrapper);
                            }

                            string alias = null;
                            if (tempWrapper.ContainsKey("alias"))
                            {
                                alias = tempWrapper["alias"].StringValue;
                            }
                            else if (tempWrapper.ContainsKey("mailNickname"))
                            {
                                alias = tempWrapper["mailNickname"].StringValue;
                            }

                            var dn = tempWrapper["dn"].StringValue;
                            var cn = Config.ADConst.GetCN(dn);
                            var ou = Config.ADConst.GetContainer(dn);

                            csentry = ExchangeUtils.CreateMailbox(
                                connectedMA,
                                connectedMA.EscapeDNComponent(cn).Concat(ou),
                                alias,
                                tempWrapper["homeMDB"].StringValue);

                            EntryWrapper cs = csentry;

                            foreach (var field in tempWrapper.Keys)
                            {
                                if (new List<string> { "dn", "alias", "mailnickname", "homemdb", "objecttype" }.Contains(field.ToLower()))
                                {
                                    continue;
                                }

                                cs[field] = tempWrapper[field];
                            }
                        }
                        else
                        {
                            csentry = mventry.ConnectedMAs[ma.Key].Connectors.StartNewConnector(rule.CSObjectType);
                            foreach (var map in rule.Rules)
                            {
                                rules.Run(map, mventry, csentry);
                            }
                        }

                        if (commit)
                        {
                            csentry.CommitNewConnector();
                        }

                        ret.Add(new ProvisionChange { Add = true, csentry = csentry, Rule = rule });
                    }
                    else
                    {
                        foreach (var csentry in mventry.ConnectedMAs[ma.Key].Connectors)
                        {
                            if (csentry.ObjectType.ToLower() == rule.CSObjectType.ToLower())
                            {
                                var currDN = csentry.DN;
                                foreach (var map in rule.Rules)
                                {
                                    rules.Run(map, mventry, csentry);
                                }

                                if (csentry.DN.ToString() != currDN.ToString())
                                {
                                    ret.Add(new ProvisionChange { Add = false, csentry = csentry, Rule = rule });
                                }
                            }
                        }
                    } // end connector count if
                }
            }

            return ret;
        } // end Provision method

        public void Dispose()
        {
            rules.Dispose();
        }
    }
}
