﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;
using Microsoft.MetadirectoryServices;

namespace SoftwareIDM.Uplift
{
    /// <summary>
    /// RuleType lists the possible kinds of rule
    /// </summary>
    public enum RuleType
    {
        Literal,     // A String literal. e.g. "Hello World."
        Integer,     // A positive or negative integer. e.g. 42
        Bool,        // Rule is literally true or false
        Function,    // A more complicated rule, represents a function to invoke
        Field,       // A field lookup on a CS or MV object.
        Concatenate, // String concatenation using the + operator
        Comparison,  // Comparison Operator
        Empty        // An empty Rule does nothing.
    }

    static class ComparisonOperators
    {
        public static AttributeValue Compare(AttributeValue left, AttributeValue right, string oper)
        {
            switch (oper)
            {
                case "<":
                    return new AttributeValue(left.CompareTo(right) < 0);
                case ">":
                    return new AttributeValue(left.CompareTo(right) > 0);
                case "<=":
                    return new AttributeValue(left.CompareTo(right) <= 0);
                case ">=":
                    return new AttributeValue(left.CompareTo(right) >= 0);
                case "==":
                    return new AttributeValue(left.Equals(right));
                default:
                    return new AttributeValue(false);
            }
        }
    }

    /// <summary>
    /// The Rule class is the core of synchronization rule parsing.
    /// There are three components. 
    ///     The constructor tokenizes the Rule input string to generate the logic.
    ///     The GetValue method passes the rule to a lazy bound AttributeValue. Rule execution is deferred until
    ///         the value is needed.
    ///     The LateBind method performs the rule logic to generate a bound AttributeValue.
    ///         The LateBind method is called by the unbound AttributeValue that was constructed with the rule
    ///         as a parameter.
    /// Rules are defined with a small DSL (Domain Specific Language) and may be 
    /// chained and nested in quite complex ways.
    /// </summary>
    public class Rule
    {
        List<Rule> parameters = new List<Rule>();
        Dictionary<string, Rule> namedParameters = new Dictionary<string, Rule>();
        public RuleType RuleType { get; set; }
        public string Token { get; set; }
        public string ComparisonOperator { get; set; }

        readonly Regex reFunc = new Regex(@"^(\w+)\(");
        readonly Regex reNamedParam = new Regex(@"^(\w+)(?:=|\s=\s)");
        readonly Regex reField = new Regex(@"^[-\w.]+");
        readonly Regex reInteger = new Regex(@"^((?:-|0x)?\d+)");
        readonly Regex reBool = new Regex(@"^(true|false)");
        readonly Regex reLit = new Regex(@"^""(.*?)(?!\\)""", RegexOptions.Multiline);
        readonly Regex reComparison = new Regex(@"^(>=|<=|>|<|==)");

        private Rule() { }

        public string Target { get; set; }

        public Rule(string rule)
        {
            Regex split = new Regex("^(.+)->(.+)$");
            string left;
            var re = split.Match(rule);
            var hold = new Rule(re.Groups[1].Value, out left);

            source = hold.source;
            target = hold.target;
            RuleType = hold.RuleType;
            Token = hold.Token;
            parameters = hold.parameters;
            namedParameters = hold.namedParameters;
            ComparisonOperator = hold.ComparisonOperator;

            if (re.Groups.Count > 2 && !String.IsNullOrEmpty(re.Groups[2].Value.Trim()))
            {
                Target = re.Groups[2].Value.Trim();
            }
        }

        /// <summary>
        /// Tokenizes the rule string into a parse tree of Rule instances.
        /// 
        /// The simple parsing algorithm consumes the rule from beginning to end, tokenizing with
        /// regular expressions (and in a couple cases String.StartsWith).
        /// In the case of the Function RuleType, Rules may be nested as function parameters.
        /// 
        /// For included available Rule Functions refer to CommonFunctions.cs
        /// </summary>
        /// <param name="rule">String to parse as a rule.</param>
        /// <param name="left">Remainder of string. Unconsumed text may be further parsed as other Rules.</param>
        public Rule(string rule, out string left)
        {
            rule = rule.Trim();

            left = "";
            if (String.IsNullOrEmpty(rule))
            {
                RuleType = RuleType.Empty;
                return;
            }

            var prevLength = rule.Length;
            while (rule.Length > 0)
            {
                if (reFunc.IsMatch(rule))
                {
                    Token = reFunc.Match(rule).Groups[1].Value;
                    RuleType = RuleType.Function;
                    rule = rule.Substring(Token.Length + 1);
                    while (!rule.StartsWith(")"))
                    {
                        if (rule.Length == 0)
                        {
                            throw new FormatException("Unable to parse rule, unmatched parenthesis.");
                        }

                        if (reNamedParam.IsMatch(rule))
                        {
                            var match = reNamedParam.Match(rule);
                            var name = match.Groups[1].Value;
                            rule = rule.Substring(match.Length);
                            namedParameters.Add(name, new Rule(rule, out rule));
                        }
                        else
                        {
                            parameters.Add(new Rule(rule, out rule));
                        }

                        if (rule.StartsWith(","))
                        {
                            rule = rule.Substring(1);
                        }

                        rule = rule.Trim();
                    }

                    rule = rule.Substring(1); // remove close paren
                }
                else if (reInteger.IsMatch(rule))
                {
                    Token = reInteger.Match(rule).Groups[1].Value;
                    RuleType = RuleType.Integer;
                    rule = rule.Substring(Token.Length);
                }
                else if (reBool.IsMatch(rule.ToLower()))
                {
                    Token = reBool.Match(rule.ToLower()).Groups[1].Value;
                    RuleType = RuleType.Bool;
                    rule = rule.Substring(Token.Length);
                }
                else if (reField.IsMatch(rule))
                {
                    Token = reField.Match(rule).Value;
                    RuleType = RuleType.Field;
                    rule = rule.Substring(Token.Length);
                }
                else if (reLit.IsMatch(rule))
                {
                    Token = reLit.Match(rule).Groups[1].Value;
                    RuleType = RuleType.Literal;
                    rule = rule.Substring(Token.Length + 2);
                    Token = Token.Replace(@"\\", @"\");
                    var specials = new Dictionary<string, string> { { "\\r", "\r" }, { "\\n", "\n" }, { "\\t", "\t" }, { "\\\"", "\"" } };
                    foreach (var pair in specials)
                    {
                        Token = Token.Replace(pair.Key, pair.Value);
                    }
                }
                else if (reComparison.IsMatch(rule))
                {
                    // copy current rule into first parameter
                    var add = new Rule
                    {
                        Token = Token,
                        RuleType = RuleType,
                        parameters = parameters,
                        namedParameters = namedParameters,
                        ComparisonOperator = ComparisonOperator
                    };

                    parameters = new List<Rule> { add };
                    Token = "";

                    var match = reComparison.Match(rule);
                    ComparisonOperator = match.Groups[1].Value;
                    Token = ComparisonOperator;
                    RuleType = RuleType.Comparison;
                    parameters.Add(new Rule(rule.Substring(match.Length), out rule));
                }
                else if (rule.StartsWith("+"))
                {
                    var add = new Rule { Token = Token, RuleType = RuleType, parameters = parameters };
                    parameters = new List<Rule> { add };
                    Token = "";

                    RuleType = RuleType.Concatenate;
                    parameters.Add(new Rule(rule.Substring(1), out rule));
                }
                else
                {
                    if (rule.Trim().Length == prevLength)
                    {
                        throw new FormatException("Unable to parse remainder of rule: " + rule);
                    }

                    left = rule.Trim();
                    return;
                }

                rule = rule.Trim();

                if (rule.Length == prevLength)
                {
                    throw new FormatException("Unable to parse remainder of rule: " + rule);
                }
            } // while
        }

        List<string> qualifyKeys = new List<string> { "mventry", "csentry", "mv", "cs" };
        object getParam(Rule parameter, EntryWrapper source, EntryWrapper target)
        {
            if (qualifyKeys.Contains(parameter.Token.ToLower()))
            {
                if (source.Type == EntryWrapperType.MVEntry &&
                    "mventry".StartsWith(parameter.Token.ToLower()) ||
                    source.Type == EntryWrapperType.CSEntry && "csentry".StartsWith(parameter.Token.ToLower()))
                {
                    return source;
                }
                else
                {
                    return target;
                }
            }
            else
            {
                return parameter.GetValue(source, target);
            }
        }

        /// <summary>
        /// Invoke a Function Rule using System.Reflection
        /// </summary>
        private AttributeValue Invoke(EntryWrapper source, EntryWrapper target)
        {
            MethodInfo method;
            try
            {
                method = RuleEngine.methods[Token];
            }
            catch (KeyNotFoundException)
            {
                throw new KeyNotFoundException(String.Format("Unable to find function {0} in loaded classes", Token));
            }

            var pi = method.GetParameters();
            var args = new List<object>();
            int positionalArg = 0;
            for (int i = 0; i < pi.Length; i++)
            {
                if (i < pi.Length - 1)
                {
                    if (namedParameters.ContainsKey(pi[i].Name))
                    {
                        args.Add(getParam(namedParameters[pi[i].Name], source, target));
                    }
                    else if (parameters.Count > positionalArg)
                    {
                        args.Add(getParam(parameters[positionalArg], source, target));
                        positionalArg++;
                    }
                    else
                    {
                        args.Add(null);
                    }
                }
                else if (pi.Last().ParameterType.IsArray)
                {
                    var array = new List<AttributeValue>();
                    while (positionalArg < parameters.Count)
                    {
                        array.Add(parameters[positionalArg].GetValue(source, target));
                        positionalArg++;
                    }

                    args.Add(array.ToArray());
                }
                else if (pi.Last().ParameterType.IsAssignableFrom(typeof(Dictionary<string, AttributeValue>)))
                {
                    var named = new Dictionary<string, AttributeValue>();
                    foreach (var key in namedParameters.Keys)
                    {
                        named[key] = namedParameters[key].GetValue(source, target);
                    }

                    args.Add(named);
                }
                else
                {
                    if (namedParameters.ContainsKey(pi[i].Name))
                    {
                        args.Add(getParam(namedParameters[pi[i].Name], source, target));
                    }
                    else if (parameters.Count > positionalArg)
                    {
                        args.Add(getParam(parameters[positionalArg], source, target));
                        positionalArg++;
                    }
                    else
                    {
                        args.Add(null);
                    }
                }
            }

            try
            {
                return (AttributeValue)method.Invoke(null, args.ToArray());
            }
            catch (TargetInvocationException e)
            {
                throw new ExtensionException(String.Format(
                    "Error executing rule {0} from {1} to {2}: {3}", Token, source.Type.ToString(), target.Type.ToString(), e.InnerException.Message), e.InnerException);
            }
            catch (TargetParameterCountException e)
            {

                throw new TargetParameterCountException(String.Format("Parameter count mismatch for {0}({1})",
                    Token,
                    String.Join(",", (from p in parameters select p.Token).ToArray())), e);
            }
        }

        EntryWrapper source;
        EntryWrapper target;

        /// <summary>
        /// Callback invoked by Rule based AttributeValue to execute the Rule and bind the value.
        /// </summary>
        public AttributeValue LateBind()
        {
            switch (RuleType)
            {
                case RuleType.Field:
                    if (source.Type == EntryWrapperType.MVEntry)
                    {
                        if (Token.ToLower().StartsWith("mventry.")) return source.Get(Token.Substring(8));
                        else if (Token.ToLower().StartsWith("mv.")) return source.Get(Token.Substring(3));
                        else if (Token.ToLower().StartsWith("csentry.")) return target.Get(Token.Substring(8));
                        else if (Token.ToLower().StartsWith("cs.")) return target.Get(Token.Substring(3));
                    }
                    else if (source.Type == EntryWrapperType.CSEntry)
                    {
                        if (Token.ToLower().StartsWith("csentry.")) return source.Get(Token.Substring(8));
                        else if (Token.ToLower().StartsWith("cs.")) return source.Get(Token.Substring(3));
                        else if (Token.ToLower().StartsWith("mventry.")) return target.Get(Token.Substring(8));
                        else if (Token.ToLower().StartsWith("mv.")) return target.Get(Token.Substring(3));
                    }

                    return source.Get(Token);
                case RuleType.Literal:
                    return new AttributeValue { StringValue = Token };
                case RuleType.Integer:
                    if (Token.StartsWith("0x"))
                    {
                        return new AttributeValue(Convert.ToInt64(Token, 16));
                    }
                    else
                    {
                        return new AttributeValue(long.Parse(Token));
                    }
                case RuleType.Bool:
                    return new AttributeValue { BooleanValue = bool.Parse(Token) };
                case RuleType.Concatenate:
                    if (parameters.Count == 2)
                    {
                        return new AttributeValue
                        {
                            StringValue = String.Join("", new string[] {
                                parameters[0].GetValue(source, target).StringValue, 
                                parameters[1].GetValue(source, target).StringValue
                            })
                        };
                    }
                    else
                    {
                        return parameters[0].GetValue(source, target);
                    }
                case RuleType.Comparison:
                    if (parameters.Count == 2)
                    {   
                        return ComparisonOperators.Compare(
                            parameters[0].GetValue(source, target),
                            parameters[1].GetValue(source, target), this.ComparisonOperator);
                    }
                    else
                    {
                        return parameters[0].GetValue(source, target);
                    }
                case RuleType.Function:
                    return Invoke(source, target);
                default:
                    return null;
            }
        }

        /// <summary>
        /// Construct an AttributeValue that's primed to bind itself when the value is accessed.
        /// </summary>
        public AttributeValue GetValue(EntryWrapper source, EntryWrapper target)
        {
            this.source = source;
            this.target = target;
            AttributeValue ret = new AttributeValue(this);
            return ret;
        }
    }
}
