﻿using System;
using System.Reflection;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

namespace SoftwareIDM.Uplift
{
    public class Change
    {
        public string Direction { get; set; }
        public string Field { get; set; }
        public AttributeValue OldValue { get; set; }
        public AttributeValue NewValue { get; set; }
    }

    public class ChangeSet : IEquatable<ChangeSet>
    {
        public Dictionary<string, AttributeValue> Hologram { get; set; }
        public Guid ObjectID { get; set; }
        public Dictionary<string, Change> Changes { get; set; }

        public ChangeSet()
        {
            Changes = new Dictionary<string, Change>();
        }

        public bool Equals(ChangeSet other)
        {
            return other.ObjectID == ObjectID;
        }
    }

    public delegate void ChangeHandler(ChangeSet Changes);

    /// <summary>
    /// This class sets up the Rule engine for parsing Attribute FlowRuleNames as 
    /// FIM Portal style synchronization rules.
    /// </summary>
    public class RuleEngine : IDisposable
    {
        /// <summary>
        /// Constructor, takes class types to draw Rule functions from.
        /// </summary>
        public RuleEngine(params Type[] types)
        {
            ImportHoloFields = new Dictionary<string, string[]>();
            ExportHoloFields = new Dictionary<string, string[]>();
            if (functions == null)
            {
                functions = new List<Type>();
            }

            functions.AddRange(types);
            var forceLicense = Config.Xml;
        }

        public static List<Type> functions;

        static Dictionary<string, MethodInfo> _methods;

        /// <summary>
        /// Dictionary of method implementations to be invoked by rules. Method references are cached
        /// in the _methods to improve performance.
        /// </summary>
        public static Dictionary<string, MethodInfo> methods
        {
            get
            {
                if (_methods == null)
                {
                    var skip = new List<string> { "Equals", "ToString", "GetHashCode", "GetType" };
                    _methods = new Dictionary<string, MethodInfo>();
                    foreach (var t in functions)
                    {
                        var mi = t.GetMethods();
                        foreach (var m in mi)
                        {
                            if (!skip.Contains(m.Name))
                            {
                                _methods[m.Name] = m;
                            }
                        }
                    }
                }
                return _methods;
            }
        }

        // Rule Cache
        Dictionary<string, Rule> rules = new Dictionary<string, Rule>();


        public Dictionary<string, string[]> ImportHoloFields { get; set; }
        public Dictionary<string, string[]> ExportHoloFields { get; set; }
        ChangeSet ImportChanges { get; set; }
        ChangeSet ExportChanges { get; set; }

        public event ChangeHandler ExportFlowEnd;
        public event ChangeHandler ImportFlowEnd;

        private Dictionary<string, AttributeValue> BuildHolo(Dictionary<string, string[]> fields, EntryWrapper entry)
        {
            var ret = new Dictionary<string, AttributeValue>();
            if (fields.ContainsKey(entry["ObjectType"].StringValue))
            {
                foreach (var s in fields[entry["ObjectType"].StringValue])
                {
                    ret.Add(s, entry[s]);
                }
            }

            return ret;
        }

        /// <summary>
        /// Handle Rule processing.
        /// </summary>
        /// <param name="rule">
        /// String with rule logic. 
        /// The first time the rule is parsed, it's cached in a dictionary.
        /// Rule takes the form Rule->target.
        /// Technically, "->target" is not part of the Rule, merely instructions on what to do
        /// with the result of the Rule's execution.
        /// </param>
        /// <param name="source">CSEntry or MVEntry inside an EntryWrapper</param>
        /// <param name="target">MVEntry or CSEntry inside an EntryWrapper</param>
        /// <returns>Whether the rule was successfully processed.</returns>
        public bool Run(string rule, EntryWrapper source, EntryWrapper target)
        {
            if (!rules.ContainsKey(rule))
            {
                if (!Regex.IsMatch(rule, "^(.+)->(.+)$"))
                {
                    return false;
                }

                rules[rule] = new Rule(rule);
            }

            var val = rules[rule].GetValue(source, target);
            var field = rules[rule].Target;

            string direction = source.Type == EntryWrapperType.CSEntry ? "Import" : "Export";
            var oid = source.Type == EntryWrapperType.MVEntry ? source.mventry.ObjectID : target.mventry.ObjectID;

            // Adjust change objects if necessary
            if (direction == "Import")
            {
                if (ImportChanges == null)
                {
                    ImportChanges = new ChangeSet { ObjectID = oid, Hologram = BuildHolo(ImportHoloFields, target) };
                }

                if (ImportChanges.ObjectID != oid)
                {
                    if (ImportFlowEnd != null)
                        ImportFlowEnd(ImportChanges);
                    ImportChanges = new ChangeSet { ObjectID = oid, Hologram = BuildHolo(ImportHoloFields, target) };
                }

                if (!val.SkipFlow && !target.Get(field).Equals(val))
                {
                    ImportChanges.Changes[field] = new Change
                    {
                        Direction = direction,
                        Field = field,
                        OldValue = target[field],
                        NewValue = val
                    };

                    // Set change
                    target.Set(field, val);
                }
            }
            else
            {
                if (ExportChanges == null)
                {
                    ExportChanges = new ChangeSet { ObjectID = oid, Hologram = BuildHolo(ExportHoloFields, target) };
                }

                if (ExportChanges.ObjectID != oid)
                {
                    if (ExportFlowEnd != null)
                        ExportFlowEnd(ExportChanges);
                    ExportChanges = new ChangeSet { ObjectID = oid, Hologram = BuildHolo(ExportHoloFields, target) };
                }

                // have to do it this way because otherwise MultiValued is not set for missing attribute
                if (!val.SkipFlow && !val.Equals(target.Get(field)))
                {
                    ExportChanges.Changes[field] = new Change
                    {
                        Direction = direction,
                        Field = field,
                        OldValue = target[field],
                        NewValue = val
                    };

                    // Set change
                    target.Set(field, val);
                }
            }

            return true;
        }

        public void Dispose()
        {
            if (ImportChanges != null)
            {
                if (ImportFlowEnd != null)
                    ImportFlowEnd(ImportChanges);
                ImportChanges = null;
            }

            if (ExportChanges != null)
            {
                if (ExportFlowEnd != null)
                    ExportFlowEnd(ExportChanges);
                ExportChanges = null;
            }
        }
    }
}
